from pathlib import Path 

PCCM_FUNC_META_KEY = "__pccm_func_meta"
PCCM_CLASS_META_KEY = "__pccm_class_meta"

PCCM_MAGIC_STRING = "// THIS FILE IS AUTO-GENERATED BY PCCM, DON'T MODIFY THIS LINE!!! md5: {}"

PCCM_INIT_DECORATOR_KEY = "__pccm_this_type"

PCCM_INLINE_LIBRARY_PATH = Path.home() / ".pccm" / "inline"

PCCM_DISABLE_CODE_CHANGE = "PCCM_DISABLE_CODE_CHANGE"